import 'dart:async';
import 'dart:io';
import 'dart:math';
import 'dart:typed_data';
import 'package:path/path.dart' as p;
import 'package:fc_native_video_thumbnail/fc_native_video_thumbnail.dart';
import 'package:file_picker/file_picker.dart';
import 'package:flutter/material.dart';
import 'package:image_picker/image_picker.dart';
import 'package:path_provider/path_provider.dart';
import 'package:tencent_cloud_chat_uikit/base_widgets/tim_ui_kit_base.dart';
import 'package:tencent_cloud_chat_uikit/base_widgets/tim_ui_kit_state.dart';
import 'package:tencent_cloud_chat_uikit/business_logic/separate_models/tui_chat_separate_view_model.dart';
import 'package:tencent_cloud_chat_uikit/business_logic/view_models/tui_chat_global_model.dart';
import 'package:tencent_cloud_chat_uikit/business_logic/view_models/tui_setting_model.dart';
import 'package:tencent_cloud_chat_uikit/data_services/core/%20tim_uikit_wide_modal_operation_key.dart';
import 'package:tencent_cloud_chat_uikit/data_services/services_locatar.dart';
import 'package:tencent_cloud_chat_uikit/tencent_cloud_chat_uikit.dart';
import 'package:tencent_cloud_chat_uikit/ui/utils/message.dart';
import 'package:tencent_cloud_chat_uikit/ui/utils/optimize_utils.dart';
import 'package:tencent_cloud_chat_uikit/ui/widgets/wide_popup.dart';
import 'package:tencent_cloud_chat_uikit/ui/utils/permission.dart';
import 'package:tencent_cloud_chat_uikit/ui/utils/platform.dart';
import 'package:tencent_cloud_chat_uikit/ui/views/TIMUIKitChat/TIMUIKitTextField/special_text/DefaultSpecialTextSpanBuilder.dart';
import 'package:tencent_cloud_chat_uikit/ui/views/TIMUIKitChat/TIMUIKitTextField/tim_uikit_emoji_panel.dart';
import 'package:tencent_cloud_chat_uikit/ui/widgets/drag_widget.dart';
import 'package:tencent_extended_text_field/extended_text_field.dart';
import 'package:universal_html/html.dart' as html;
import 'package:wechat_assets_picker/wechat_assets_picker.dart';

class WideControlBarItem {
  final String item;
  final IconData? icon;
  final String? imgPath;
  final Color? color;
  final ValueChanged<Offset?> onClick;
  final String? showName;
  final double? size;

  WideControlBarItem(
      {required this.item,
      this.icon,
      this.color,
      this.imgPath,
      required this.onClick,
      this.showName,
      this.size})
      : assert(icon != null || imgPath != null);
}

class TIMUIKitTextFieldLayoutWide extends StatefulWidget {
  /// sticker panel customization
  final CustomStickerPanel? customStickerPanel;
  final VoidCallback onEmojiSubmitted;
  final Function(int, String) onCustomEmojiFaceSubmitted;
  final Function(String, bool) handleSendEditStatus;
  final VoidCallback backSpaceText;
  final ValueChanged<String> addStickerToText;

  final ValueChanged<String> handleAtText;

  /// Whether to use the default emoji
  final bool isUseDefaultEmoji;

  final TUIChatSeparateViewModel model;

  /// background color
  final Color? backgroundColor;

  /// control input field behavior
  final TIMUIKitInputTextFieldController? controller;

  /// config for more panel
  final MorePanelConfig? morePanelConfig;

  final String languageType;

  final TextEditingController textEditingController;

  /// conversation id
  final String conversationID;

  /// conversation type
  final ConvType conversationType;

  final FocusNode focusNode;

  /// show more panel
  final bool showMorePanel;

  /// hint text for textField widget
  final String? hintText;

  final int? currentCursor;

  final ValueChanged<int?> setCurrentCursor;

  final VoidCallback onCursorChange;

  /// show send audio icon
  final bool showSendAudio;

  final VoidCallback handleSoftKeyBoardDelete;

  /// on text changed
  final void Function(String)? onChanged;

  final V2TimMessage? repliedMessage;

  /// show send emoji icon
  final bool showSendEmoji;

  final String? forbiddenText;

  final VoidCallback onSubmitted;

  final VoidCallback goDownBottom;

  final List customEmojiStickerList;

  /// Conversation need search
  final V2TimConversation currentConversation;

  const TIMUIKitTextFieldLayoutWide(
      {Key? key,
      this.customStickerPanel,
      required this.onEmojiSubmitted,
      required this.onCustomEmojiFaceSubmitted,
      required this.backSpaceText,
      required this.addStickerToText,
      required this.isUseDefaultEmoji,
      required this.languageType,
      required this.textEditingController,
      this.morePanelConfig,
      required this.conversationID,
      required this.conversationType,
      required this.focusNode,
      this.currentCursor,
      required this.setCurrentCursor,
      required this.onCursorChange,
      required this.model,
      this.backgroundColor,
      this.onChanged,
      required this.handleSendEditStatus,
      required this.handleAtText,
      required this.handleSoftKeyBoardDelete,
      this.repliedMessage,
      this.forbiddenText,
      required this.onSubmitted,
      required this.goDownBottom,
      required this.showSendAudio,
      required this.showSendEmoji,
      required this.showMorePanel,
      this.hintText,
      required this.customEmojiStickerList,
      this.controller,
      required this.currentConversation})
      : super(key: key);

  @override
  State<TIMUIKitTextFieldLayoutWide> createState() =>
      _TIMUIKitTextFieldLayoutWideState();
}

class _TIMUIKitTextFieldLayoutWideState
    extends TIMUIKitState<TIMUIKitTextFieldLayoutWide> {
  final TUISettingModel settingModel = serviceLocator<TUISettingModel>();
  OverlayEntry? entry;
  final ImagePicker _picker = ImagePicker();
  Uint8List? fileContent;
  String? fileName;
  File? tempFile;
  Function? setKeyboardHeight;
  double? bottomPadding;
  late ScrollController _scrollController;

  @override
  void initState() {
    super.initState();
    if (widget.controller != null) {
      widget.controller?.addListener(() {
        final actionType = widget.controller?.actionType;
        if (actionType == ActionType.hideAllPanel) {
          hideAllPanel();
        }
      });
    }
    widget.focusNode.requestFocus();
    _scrollController = ScrollController();
  }

  hideAllPanel() {
    widget.focusNode.unfocus();
    widget.currentCursor == null;
  }

  _debounce(
    Function(String text) fun, [
    Duration delay = const Duration(milliseconds: 30),
  ]) {
    Timer? timer;
    return (String text) {
      if (timer != null) {
        timer?.cancel();
      }

      timer = Timer(delay, () {
        fun(text);
      });
    };
  }

  _buildRepliedMessage(V2TimMessage? repliedMessage) {
    final haveRepliedMessage = repliedMessage != null;
    if (haveRepliedMessage) {
      final text =
          "${MessageUtils.getDisplayName(widget.model.repliedMessage!)}:${widget.model.abstractMessageBuilder != null ? widget.model.abstractMessageBuilder!(widget.model.repliedMessage!) : MessageUtils.getAbstractMessageAsync(widget.model.repliedMessage!, widget.model.groupMemberList ?? [])}";
      return Container(
        color: widget.backgroundColor ?? hexToColor("f5f5f6"),
        alignment: Alignment.centerLeft,
        padding: const EdgeInsets.symmetric(vertical: 10, horizontal: 16),
        child: Row(
          crossAxisAlignment: CrossAxisAlignment.center,
          mainAxisAlignment: MainAxisAlignment.spaceBetween,
          children: [
            Expanded(
              child: Text(
                text,
                softWrap: true,
                maxLines: 3,
                overflow: TextOverflow.ellipsis,
                style: TextStyle(color: hexToColor("8f959e"), fontSize: 14),
              ),
            ),
            const SizedBox(
              width: 16,
            ),
            InkWell(
              onTap: () {
                widget.model.repliedMessage = null;
              },
              child: Icon(Icons.clear, color: hexToColor("8f959e"), size: 18),
            )
          ],
        ),
      );
    }
    return Container();
  }

  _sendEmoji(Offset? offset, TUITheme theme) {
    widget.onCursorChange();
    if (entry != null) {
      entry?.remove();
      entry = null;
    } else {
      entry = OverlayEntry(builder: (BuildContext context) {
        return TUIKitDragArea(
            closeFun: () {
              if (entry != null) {
                entry?.remove();
                entry = null;
              }
            },
            initOffset: offset ??
                Offset(MediaQuery.of(context).size.height * 0.5 + 20,
                    MediaQuery.of(context).size.height * 0.5 - 100),
            child: Container(
              decoration: BoxDecoration(
                borderRadius: const BorderRadius.all(Radius.circular(8)),
                color: theme.wideBackgroundColor,
                border: Border.all(
                  width: 2,
                  color: theme.weakBackgroundColor ?? const Color(0xFFbebebe),
                ),
                boxShadow: const [
                  BoxShadow(
                    color: Color(0xFFbebebe),
                    offset: Offset(5, 5),
                    blurRadius: 10,
                    spreadRadius: 1,
                  ),
                ],
              ),
              child: Container(
                child: widget.customStickerPanel != null
                    ? widget.customStickerPanel!(
                        height: 400,
                        width: 350,
                        sendTextMessage: () {
                          widget.onEmojiSubmitted();
                        },
                        sendFaceMessage: widget.onCustomEmojiFaceSubmitted,
                        deleteText: () {
                          widget.backSpaceText();
                        },
                        addText: (int unicode) {
                          final newText = String.fromCharCode(unicode);
                          widget.addStickerToText(newText);
                        },
                        addCustomEmojiText: ((String singleEmojiName) {
                          String? emojiName = singleEmojiName.split('.png')[0];
                          if (widget.isUseDefaultEmoji &&
                              widget.languageType == 'zh' &&
                              ConstData.emojiMapList[emojiName] != null &&
                              ConstData.emojiMapList[emojiName] != '') {
                            emojiName = ConstData.emojiMapList[emojiName];
                          }
                          final newText = '[$emojiName]';
                          widget.addStickerToText(newText);
                        }),
                        defaultCustomEmojiStickerList:
                            widget.isUseDefaultEmoji ? ConstData.emojiList : [])
                    : EmojiPanel(onTapEmoji: (unicode) {
                        final newText = String.fromCharCode(unicode);
                        widget.addStickerToText(newText);
                      }, onSubmitted: () {
                        widget.onEmojiSubmitted();
                      }, delete: () {
                        widget.backSpaceText();
                      }),
              ),
            ));
      });
      Overlay.of(context).insert(entry!);
    }
  }

  _addGreyOverlay() {
    if (entry != null) {
      _removeOverlay();
      return;
    } else {
      entry = OverlayEntry(builder: (BuildContext context) {
        return Container(
          color: const Color(0x7F000000),
        );
      });
      Overlay.of(context).insert(entry!);
    }
  }

  _removeOverlay() {
    entry!.remove();
    entry = null;
  }

  _sendFile(
    TUIChatSeparateViewModel model,
    TUITheme theme,
  ) async {
    if (PlatformUtils().isMobile &&
        !await Permissions.checkPermission(
          context,
          Permission.storage.value,
          theme,
        )) {
      return;
    }
    if(!PlatformUtils().isWeb){
      _addGreyOverlay();
    }
    try {
      final convID = widget.conversationID;
      final convType = widget.conversationType;
      FilePickerResult? result = await FilePicker.platform.pickFiles();
      _removeOverlay();
      if (result != null && result.files.isNotEmpty) {
        if (PlatformUtils().isWeb) {
          html.Node? inputElem;
          inputElem = html.document
              .getElementById("__file_picker_web-file-input")
              ?.querySelector("input");
          fileName = result.files.single.name;

          MessageUtils.handleMessageError(
              model.sendFileMessage(
                  inputElement: inputElem,
                  fileName: fileName,
                  convID: convID,
                  convType: convType),
              context);
          return;
        }

        String? option2 = result.files.single.path ?? "";
        print(TIM_t_para("选择成功{{option2}}", "选择成功$option2")(option2: option2));
        File file = File(result.files.single.path!);
        final int size = file.lengthSync();
        final String savePath = file.path;

        MessageUtils.handleMessageError(
            model.sendFileMessage(
                filePath: savePath,
                size: size,
                convID: convID,
                convType: convType),
            context);
      } else {
        throw NullThrownError();
      }
    } catch (e) {
      print("_sendFileErr: ${e.toString()}");
    }
  }

  List<Widget> generateBarIcons(
      List<WideControlBarItem> items, TUITheme theme) {
    return items.map((e) {
      final GlobalKey key = GlobalKey();
      return Container(
        margin: const EdgeInsets.only(right: 6),
        child: InkWell(
          onTap: () {
            final alignBox =
                key.currentContext?.findRenderObject() as RenderBox?;
            var offset = alignBox?.localToGlobal(Offset.zero);
            final double? dx = (offset?.dx != null) ? offset!.dx : null;
            final double? dy =
                (offset?.dy != null && alignBox?.size.height != null)
                    ? offset!.dy - 420
                    : null;
            e.onClick((dx != null && dy != null) ? Offset(dx, dy) : null);
          },
          child: Tooltip(
            preferBelow: false,
            textStyle: TextStyle(fontSize: 12, color: theme.white),
            message: e.showName,
            child: Container(
                decoration:
                    BoxDecoration(borderRadius: BorderRadius.circular(2)),
                padding: const EdgeInsets.all(4),
                child: e.imgPath != null
                    ? Image.asset(
                        e.imgPath!,
                        key: key,
                        width: e.size ?? 20,
                        height: e.size ?? 20,
                      )
                    : Icon(e.icon,
                        key: key,
                        color: e.color ?? hexToColor("646a73"),
                        size: e.size ?? 20)),
          ),
        ),
      );
    }).toList();
  }

  _sendImageFileOnWeb(TUIChatSeparateViewModel model) async {
    try {
      final pickedFile = await _picker.pickImage(source: ImageSource.gallery);
      final imageContent = await pickedFile!.readAsBytes();
      fileName = pickedFile.name;
      tempFile = File(pickedFile.path);
      fileContent = imageContent;

      html.Node? inputElem;
      inputElem = html.document
          .getElementById("__image_picker_web-file-input")
          ?.querySelector("input");
      final convID = widget.conversationID;
      final convType = widget.conversationType;
      MessageUtils.handleMessageError(
          model.sendImageMessage(
              inputElement: inputElem,
              imagePath: tempFile?.path,
              convID: convID,
              convType: convType),
          context);
    } catch (e) {
      print("_sendFileErr: ${e.toString()}");
    }
  }

  _sendVideoFileOnWeb(TUIChatSeparateViewModel model) async {
    try {
      final pickedFile = await _picker.pickVideo(source: ImageSource.gallery);
      final videoContent = await pickedFile!.readAsBytes();
      fileName = pickedFile.name;
      tempFile = File(pickedFile.path);
      fileContent = videoContent;

      if (fileName!.split(".")[fileName!.split(".").length - 1] != "mp4") {
        onTIMCallback(TIMCallback(
            type: TIMCallbackType.INFO,
            infoRecommendText: TIM_t("视频消息仅限 mp4 格式"),
            infoCode: 6660412));
        return;
      }

      html.Node? inputElem;
      inputElem = html.document
          .getElementById("__image_picker_web-file-input")
          ?.querySelector("input");
      final convID = widget.conversationID;
      final convType = widget.conversationType;
      MessageUtils.handleMessageError(
          model.sendVideoMessage(
              inputElement: inputElem,
              videoPath: tempFile?.path,
              convID: convID,
              convType: convType),
          context);
    } catch (e) {
      print("_sendFileErr: ${e.toString()}");
    }
  }

  _sendVideoMessage(AssetEntity asset, TUIChatSeparateViewModel model) async {
    final plugin = FcNativeVideoThumbnail();
    final originFile = await asset.originFile;
    final size = await originFile!.length();
    if (size >= 104857600) {
      onTIMCallback(TIMCallback(
          type: TIMCallbackType.INFO,
          infoRecommendText: TIM_t("发送失败,视频不能大于100MB"),
          infoCode: 6660405));
      return;
    }

    final duration = asset.videoDuration.inSeconds;
    final filePath = originFile.path;
    final convID = widget.conversationID;
    final convType = widget.conversationType;

    String tempPath = (await getTemporaryDirectory()).path +
        p.extension(originFile.path, 3) +
        ".jpeg";

    await plugin.getVideoThumbnail(
      srcFile: originFile.path,
      keepAspectRatio: true,
      destFile: tempPath,
      format: 'jpeg',
      width: 128,
      quality: 100,
      height: 128,
    );
    MessageUtils.handleMessageError(
        model.sendVideoMessage(
            videoPath: filePath,
            duration: duration,
            snapshotPath: tempPath ?? '',
            convID: convID,
            convType: convType),
        context);
  }

  _sendMediaMessage(
      TUIChatSeparateViewModel model, TUITheme theme, FileType fileType) async {
    try {
      final bool isAndroid = PlatformUtils().isAndroid;
      if (PlatformUtils().isMobile &&
          !await Permissions.checkPermission(
            context,
            isAndroid ? Permission.storage.value : Permission.photos.value,
            theme,
          )) {
        return;
      }
      final convID = widget.conversationID;
      final convType = widget.conversationType;

      if (PlatformUtils().isMobile) {
        final pickedAssets = await AssetPicker.pickAssets(context);

        if (pickedAssets != null) {
          for (var asset in pickedAssets) {
            final originFile = await asset.originFile;
            final filePath = originFile?.path;
            final type = asset.type;
            if (filePath != null) {
              if (type == AssetType.image) {
                MessageUtils.handleMessageError(
                    model.sendImageMessage(
                        imagePath: filePath,
                        convID: convID,
                        convType: convType),
                    context);
              }

              if (type == AssetType.video) {
                _sendVideoMessage(asset, model);
              }
            }
          }
        }
      } else {
        final plugin = FcNativeVideoThumbnail();
        _addGreyOverlay();
        FilePickerResult? result =
            await FilePicker.platform.pickFiles(type: fileType);
        _removeOverlay();
        if (result != null && result.files.isNotEmpty) {
          File file = File(result.files.single.path!);
          final String savePath = file.path;
          final String type = TencentUtils.getFileType(
                  (savePath.split(".")[savePath.split(".").length - 1])
                      .toLowerCase())
              .split("/")[0];

          if (type == "image") {
            MessageUtils.handleMessageError(
                model.sendImageMessage(
                    imagePath: savePath, convID: convID, convType: convType),
                context);
          } else if (type == "video") {
            String tempPath = (await getTemporaryDirectory()).path +
                p.basename(savePath) +
                ".jpeg";
            await plugin.getVideoThumbnail(
              srcFile: savePath,
              keepAspectRatio: true,
              destFile: tempPath,
              format: 'jpeg',
              width: 128,
              quality: 100,
              height: 128,
            );
            MessageUtils.handleMessageError(
                model.sendVideoMessage(
                    videoPath: savePath,
                    convID: convID,
                    convType: convType,
                    snapshotPath: tempPath),
                context);
          }
        } else {
          throw NullThrownError();
        }
      }
    } catch (err) {
      print("err: $err");
    }
  }

  List<Widget> generateControlBar(
      TUIChatSeparateViewModel model, TUITheme theme) {
    final List<WideControlBarItem> itemsList =
        model.chatConfig.wideControlBar ??
            [
              WideControlBarItem(
                  item: "face",
                  showName: TIM_t("表情"),
                  onClick: (offset) {
                    _sendEmoji(offset, theme);
                  },
                  icon: Icons.mood),
              WideControlBarItem(
                  item: "file",
                  showName: TIM_t("文件"),
                  onClick: (offset) {
                    _sendFile(widget.model, theme);
                  },
                  icon: Icons.file_copy_outlined),
              WideControlBarItem(
                  item: "photo",
                  showName: TIM_t("图片"),
                  onClick: (offset) {
                    if (PlatformUtils().isWeb) {
                      _sendImageFileOnWeb(widget.model);
                    } else {
                      _sendMediaMessage(widget.model, theme, FileType.image);
                    }
                  },
                  icon: Icons.image_outlined),
              WideControlBarItem(
                  item: "video",
                  showName: TIM_t("视频"),
                  onClick: (offset) {
                    if (PlatformUtils().isWeb) {
                      _sendVideoFileOnWeb(widget.model);
                    } else {
                      _sendMediaMessage(widget.model, theme, FileType.video);
                    }
                  },
                  icon: Icons.video_library_outlined),
              WideControlBarItem(
                  item: "history",
                  showName: TIM_t("消息历史"),
                  onClick: (offset) {
                    TUIKitWidePopup.showPopupWindow(
                        operationKey: TUIKitWideModalOperationKey.chatHistory,
                        context: context,
                        width: MediaQuery.of(context).size.width * 0.5,
                        height: MediaQuery.of(context).size.width * 0.5,
                        child: (onClose) => TIMUIKitSearchMsgDetail(
                              currentConversation: widget.currentConversation,
                              keyword: '',
                              initMessageList: widget.model
                                  .getOriginMessageList()
                                  .getRange(
                                      0,
                                      min(
                                          widget.model
                                              .getOriginMessageList()
                                              .length,
                                          100))
                                  .toList(),
                              onTapConversation:
                                  (V2TimConversation conversation,
                                      V2TimMessage? message) {},
                            ),
                        theme: theme);
                  },
                  icon: Icons.chat_outlined),
            ];
    return generateBarIcons(itemsList, theme);
  }

  @override
  Widget tuiBuild(BuildContext context, TUIKitBuildValue value) {
    final theme = value.theme;
    bool isNeedClearRepliedMessage = false;

    setKeyboardHeight ??= OptimizeUtils.debounce((height) {
      settingModel.keyboardHeight = height;
    }, const Duration(seconds: 1));

    final debounceFunc = _debounce((value) {
      if (widget.onChanged != null) {
        widget.onChanged!(value);
      }
      widget.handleAtText(value);
      widget.handleSendEditStatus(value, true);
      final isEmpty = value.isEmpty;
      if (isEmpty) {
        widget.handleSoftKeyBoardDelete();
      }
    }, const Duration(milliseconds: 80));

    final MediaQueryData data = MediaQuery.of(context);
    EdgeInsets padding = data.padding;
    if (bottomPadding == null || padding.bottom > bottomPadding!) {
      bottomPadding = padding.bottom;
    }

    return Container(
      color: widget.backgroundColor,
      child: Column(
        children: [
          SizedBox(
              height: 1,
              child:
                  Container(color: theme.weakDividerColor ?? Colors.black12)),
          Container(
            padding: const EdgeInsets.symmetric(vertical: 4, horizontal: 12),
            child: Row(
              mainAxisAlignment: MainAxisAlignment.start,
              children: generateControlBar(widget.model, theme),
            ),
          ),
          _buildRepliedMessage(widget.repliedMessage),
          Container(
            padding: const EdgeInsets.symmetric(vertical: 6, horizontal: 6),
            constraints: const BoxConstraints(minHeight: 50),
            child: Row(
              children: [
                if (widget.forbiddenText != null)
                  Expanded(
                      child: Container(
                    height: 35,
                    color: theme.weakBackgroundColor,
                    alignment: Alignment.center,
                    child: Text(
                      TIM_t(widget.forbiddenText!),
                      textAlign: TextAlign.center,
                      style: TextStyle(
                        fontWeight: FontWeight.bold,
                        fontSize: 16,
                        color: theme.weakTextColor,
                      ),
                    ),
                  )),
                if (widget.forbiddenText == null)
                  Expanded(
                    child: Scrollbar(
                      controller: _scrollController,
                      child: ExtendedTextField(
                          autofocus: true,
                          maxLines: 6,
                          minLines: 6,
                          focusNode: widget.focusNode,
                          onChanged: debounceFunc,
                          keyboardType: TextInputType.multiline,
                          textInputAction: PlatformUtils().isAndroid
                              ? TextInputAction.newline
                              : TextInputAction.send,
                          onEditingComplete: () {
                            widget.onSubmitted();
                          },
                          textAlignVertical: TextAlignVertical.top,
                          style: const TextStyle(fontSize: 14),
                          decoration: InputDecoration(
                            hoverColor: hexToColor("fafafa"),
                            border: InputBorder.none,
                            hintStyle: const TextStyle(
                              color: Color(0xffAEA4A3),
                            ),
                            fillColor: hexToColor("fafafa"),
                            filled: true,
                            isDense: true,
                            hintText: widget.hintText ?? '',
                          ),
                          controller: widget.textEditingController,
                          specialTextSpanBuilder: PlatformUtils().isWeb
                              ? null
                              : DefaultSpecialTextSpanBuilder(
                            isUseDefaultEmoji: widget.isUseDefaultEmoji,
                            customEmojiStickerList:
                            widget.customEmojiStickerList,
                            showAtBackground: true,
                          )),
                    ),
                  ),
              ],
            ),
          ),
        ],
      ),
    );
  }
}
